<?php

namespace Drupal\deprecation_status\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\deprecation_status\DataSource;

class ErrorsForm extends FormBase {

  /**
   * Stored sorting criteria.
   *
   * @type int
   */
  protected $column;

  /**
   * Stored sorting criteria ('asc' or 'desc').
   *
   * @type string
   */
  protected $sort;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'deprecation_status_errors_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['#cache']['contexts'][] = 'session';
    $form['#attached']['library'][] = 'deprecation_status/deprecation_status.lists';

    $form['#method'] = 'get';
    $form['filters'] = [
      '#type' => 'details',
      '#title' => $this->t('Filter errors'),
      '#attributes' => ['class' => ['container-inline']],
      '#open' => TRUE,
    ];
    $form['filters']['message'] = [
      '#title' => 'Error',
      '#type' => 'textfield',
      '#default_value' => @$_GET['message'],
      '#size' => 20,
    ];
    $form['filters']['project'] = [
      '#title' => 'Project',
      '#type' => 'textfield',
      '#default_value' => @$_GET['project'],
      '#size' => 20,
    ];
    $form['filters']['category'] = [
      '#title' => 'Category',
      '#type' => 'select',
      '#options' => ['Fix now' => 'Fix now', 'Once 8.7 unsupported' => 'Once 8.7 unsupported', 'Uncategorized' => 'Uncategorized', 'Parse error' => 'Parse error'],
      '#default_value' => @$_GET['category'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['group'] = [
      '#title' => 'Group',
      '#type' => 'select',
      '#options' => [],
      '#default_value' => @$_GET['group'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    foreach([50, 100, 200, 500, 1000] as $number) {
      $form['filters']['group']['#options'][$number] = '<= ' . $number;
    }
    $form['filters']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    // Read all errors into one big array.
    $errors = [];
    $file = fopen(DataSource::getFullPath('error_occurence_results.csv'), 'r');
    $i = 0;
    while ($line = fgetcsv($file, 0, ";")) {
      // Skip the header row.
      if ($i > 0) {
        switch($line[3]) {
          case 'now':
            $line[3] = 'Fix now';
            break;
          case '8.8':
            $line[3] = 'Once 8.7 unsupported';
            break;
          case 'parse':
            $line[3] = 'Parse error';
            break;
          case 'n/a':
            $line[3] = 'Uncategorized';
            break;
        }
        $errors[] = $line;
      }
      $i++;
    }
    fclose($file);

    // Set up columns for the table.
    $columns = [
      'total' => 'Total occurrences',
      'projects' => 'Total affected',
      'top' => 'Topmost group',
      'category' => 'Category',
      'message' => 'Message',
    ];
    if (!empty($_GET['project']) || !empty($_GET['group'])) {
      $columns['filtered_total'] = 'Filtered occurrences';
      $columns['filtered_projects'] = 'Filtered affected';
    }
    // Add sorting information to columns.
    $i = 0;
    foreach ($columns as $name => &$column) {
      $column = ['data' => $column, 'field' => $i];
      $i++;
    }
    // Sort by total descending initially.
    $columns['total']['sort'] = 'desc';

    $errors = $this->tableFilter($errors);

    // Sort the data array based on query tablesort information.
    $order = tablesort_get_order($columns);
    $this->column = $order['sql'] + ($order['sql'] > 4 ? 2 : 0);
    $this->sort = tablesort_get_sort($columns);
    uasort($errors, [$this, 'tableSort']);

    // Display the resulting data.
    $form['table'] = [
      '#type' => 'table',
      '#header' => $columns,
      '#empty' => t('No matching errors found.'),
    ];

    // Page the results based on query information.
    $page = pager_find_page();
    $num_per_page = 30;
    pager_default_initialize(count($errors), $num_per_page);
    $offset = $num_per_page * $page;
    $errors_page = array_slice($errors, $offset, $num_per_page);

    // Key results based on header key names.
    $keys = array_keys($columns);
    foreach ($errors_page as $error) {
      if ($error[1] == 1) {
        list($project) = explode('[', $error[5]);
        $error[1] = [
          'data' => [
            '#type' => 'link',
            '#title' => $project,
            '#url' => Url::fromRoute('deprecation_status.errors_form', [], ['query' => ['project' => $project]]),
          ]
        ];
      }
      else {
        $error[1] = [
          'data' => [
            '#type' => 'link',
            '#title' => $this->formatPlural($error[1], '@count project', '@count projects'),
            '#url' => Url::fromRoute('deprecation_status.error', ['error_index' => $error[6]]),
            '#attributes' => ['class' => ['button']],
          ]
        ];
      }
      $error_class = 'warning';
      if ($error[3] == 'Fix now') {
        $error_class = 'error';
      }
      elseif ($error[3] == 'Uncategorized') {
        $error_class = 'unknown';
      }

      // Link deprecated function names assuming they are core functions.
      $api_version = '8.8.x';
      $api_link = 'https://api.drupal.org/api/drupal/' . $api_version . '/search/';
      $formatted_error = preg_replace('!deprecated function ([^(]+)\(\)!', 'deprecated function <a target="_blank" href="' . $api_link . '\1">\1()</a>', $error[4]);

      // Replace deprecated class links.
      if (preg_match_all('!([cC]lass|trait|[uU]se) (Drupal\\\\[a-z_0-9A-Z\\\\]+)( |\.|$|:)!', $formatted_error, $found)) {
        foreach($found[2] as $found_item) {
        if (preg_match('!Drupal\\\\([a-z_0-9A-Z]+)\\\\(.+)$!', $found_item, $namespace)) {

          $path_parts = explode('\\', $namespace[2]);
          $class = array_pop($path_parts);
          if (in_array($namespace[1], ['Component', 'Core'])) {
            $class_file = 'core!lib!Drupal!' . $namespace[1];
          }
          elseif (in_array($namespace[1], ['KernelTests', 'FunctionalTests', 'FunctionalJavascriptTests'])) {
            $class_file = 'core!tests!Drupal!' . $namespace[1];
          }
          elseif ($namespace[1] == 'Tests') {
            // If the test was in a module, use that path.
            if (preg_match('!^([a-z0-9_]+)\\\\!', $namespace[2], $module)) {
              // Remove module name from array.
              array_shift($path_parts);
              $class_file = 'core!modules!' . $module[1] . '!tests!src';
            }
            else {
              // If the test was global, use that.
              $class_file = 'core!tests!Drupal!' . $namespace[1];
            }
          }
          else {
            $class_file = 'core!modules!' . $namespace[1] . '!src';
          }

          if (count($path_parts)) {
            $class_file .= '!' . join('!', $path_parts);
          }

          $class_file .= '!' . $class . '.php';
          $type = 'class';
          if (strpos($class_file, 'Interface')) {
            $type = 'interface';
          }
          elseif (strpos($class_file, 'Trait')) {
            $type = 'trait';
          }
          $api_link = 'https://api.drupal.org/api/drupal/' . $class_file . '/' . $type . '/' . $class . '/' . $api_version;
          $formatted_error = str_replace($found_item, '<a target="_blank" href="' . $api_link . '">' . $found_item . '</a>', $formatted_error);
        }
        }
      }

      // Convert remaining links.
      $formatted_error = preg_replace('!(https://www\.drupal\.org/[a-zA-Z/0-9]+)(.| |$)!', '<a target="_blank" href="\1">\1</a>\2', $formatted_error);

      // Allow error messages to wrap.
      $formatted_error = str_replace('\\', '&#8203;\\&#8203;', $formatted_error);

      $row = [
        '#attributes' => ['class' => ['item-' . $error_class]],
        'total' => ['#type' => 'markup', '#markup' => $error[0]],
        'projects' => $error[1],
        'top' => ['#type' => 'markup', '#markup' => $error[2]],
        'category' => ['#type' => 'markup', '#markup' => $error[3], '#wrapper_attributes' => ['class' => ['category']]],
        'message' => ['#type' => 'markup', '#markup' => $formatted_error],
      ];
      if (!empty($_GET['project']) || !empty($_GET['group'])) {
        $row['filtered_total'] = ['#type' => 'markup', '#markup' => $error[7]];
        $query = [];
        if (!empty($_GET['project'])) {
          $query['project'] = $_GET['project'];
        }
        if (!empty($_GET['group'])) {
          $query['group'] = $_GET['group'];
        }
        $row['filtered_projects'] = [
          'data' => [
            '#type' => 'link',
            '#title' => $this->formatPlural($error[8], '@count project', '@count projects'),
            '#url' => Url::fromRoute('deprecation_status.error', ['error_index' => $error[6]], ['query' => $query]),
            '#attributes' => ['class' => ['button']],
          ]
        ];
      }
      $form['table'][] = $row;
    }

    $form['pager'] = array(
      '#type' => 'pager'
    );

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {}

  /**
   * Custom table sorting based on stored sorting criteria.
   */
  protected function tableSort($a, $b) {
    if ($a[$this->column] == $b[$this->column]) {
      return 0;
    }
    $cmp = ($a[$this->column] < $b[$this->column]) ? -1 : 1;
    return ($this->sort == 'desc') ? $cmp * -1 : $cmp;
  }

  /**
   * Custom table filter based on form values.
   */
  protected function tableFilter($errors) {
    $keys = ['message', 'project', 'category', 'group'];
    foreach ($keys as $key) {
      $$key = @$_GET[$key];
    }

    $filtered = [];
    foreach ($errors as $index => $error) {
      // Assume this row is a match.
      $match = TRUE;
      $counts = [];

      // Go through each filter and if the row is still a match, see if this
      // filter will invalidate it.
      if (!empty($message) && $match) {
        $match = (strpos($error[4], $message) !== FALSE);
      }
      if (!empty($category) && $match) {
        $match = ($error[3] == $category);
      }
      if (!empty($group) && $match) {
        if ($error[2] <= $group) {
          // If the group was a match for the topmost group, extract affected projects and counts.
          $projects = array_slice($error, 5);
          foreach ($projects as $item) {
            if ($group == 50 && preg_match('!\[50\] \((\d+)\)$!', $item, $found)) {
              $counts[$item] = $found[1];
            }
            if ($group == 100 && preg_match('!\[(50|100)\] \((\d+)\)$!', $item, $found)) {
              $counts[$item] = $found[2];
            }
            if ($group == 200 && preg_match('!\[(50|100|200)\] \((\d+)\)$!', $item, $found)) {
              $counts[$item] = $found[2];
            }
            if ($group == 500 && preg_match('!\[(50|100|200|500)\] \((\d+)\)$!', $item, $found)) {
              $counts[$item] = $found[2];
            }
            if ($group == 1000 && preg_match('!\[(50|100|200|500|1000)\] \((\d+)\)$!', $item, $found)) {
              $counts[$item] = $found[2];
            }
          }
        }
        else {
          // The group was not a match.
          $match = FALSE;
        }
      }
      if (!empty($project) && $match) {
        $projects = array_slice($error, 5);
        // Assume match will become false and only set to TRUE if explicit.
        $match = FALSE;
        foreach ($projects as $item) {
          if (strpos($item, $project) !== FALSE) {
            if (preg_match('!\((\d+)\)$!', $item, $found)) {
              $counts[$item] = $found[1];
            }
            $match = TRUE;
          }
        }
      }

      // If the row was still a match, continue.
      if ($match) {
        $error = array_slice($error, 0, 6);
        $error[] = $index;
        if (!empty($project) || !empty($group)) {
          $error[] = (int) array_sum($counts);
          $error[] = (int) count($counts);
        }
        $filtered[] = $error;
      }
    }

    return $filtered;
  }
}
