<?php

namespace Drupal\deprecation_status\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\deprecation_status\DataSource;

class ProjectsForm extends FormBase {

  /**
   * Stored sorting criteria.
   *
   * @type int
   */
  protected $column;

  /**
   * Stored sorting criteria ('asc' or 'desc').
   *
   * @type string
   */
  protected $sort;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'deprecation_status_projects_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['#cache']['contexts'][] = 'session';
    $form['#attached']['library'][] = 'deprecation_status/deprecation_status.lists';

    $form['#method'] = 'get';
    $form['filters'] = [
      '#type' => 'details',
      '#title' => $this->t('Filter projects'),
      '#attributes' => ['class' => ['container-inline']],
      '#open' => TRUE,
    ];
    $form['filters']['name'] = [
      '#title' => 'Name',
      '#type' => 'textfield',
      '#default_value' => @$_GET['name'],
      '#size' => 20,
    ];
    $form['filters']['result'] = [
      '#title' => 'Result',
      '#type' => 'select',
      '#options' => [
        'No problems found' => 'No problems found',
        'All problems fixable now' => 'All problems fixable now',
        'Some problems fixable now' => 'Some problems fixable now',
        'Needs manual review' => 'Needs manual review',
        'Only problems fixable later' => 'Only problems fixable later',
        'No results' => 'No results',
      ],
      '#default_value' => @$_GET['result'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['error'] = [
      '#title' => 'Error',
      '#type' => 'select',
      '#options' => [
        'Environment' => 'Environment',
        'Composer' => 'Composer',
      ],
      '#default_value' => @$_GET['error'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['type'] = [
      '#title' => 'Type',
      '#type' => 'select',
      '#options' => [
        'Module' => 'Module',
        'Theme' => 'Theme',
        'Theme_engine' => 'Theme_engine',
        'Distribution' => 'Distribution',
        'Core' => 'Core',
      ],
      '#default_value' => @$_GET['type'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['group'] = [
      '#title' => 'Group',
      '#type' => 'select',
      '#options' => [],
      '#default_value' => @$_GET['group'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    foreach([50, 100, 200, 500, 1000] as $number) {
      $form['filters']['group']['#options'][$number] = '<= ' . $number;
    }
    $form['filters']['plan'] = [
      '#title' => 'Drupal 9 plan',
      '#type' => 'select',
      '#options' => ['none' => 'None', 'specified' => 'Specified'],
      '#default_value' => @$_GET['plan'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    // Read all projects into one big array.
    $projects = [];
    $file = fopen(DataSource::getFullPath('project_results.csv'), 'r');
    $i = 0;
    while ($line = fgetcsv($file, 0, ";")) {
      // Skip the header row.
      if ($i > 0) {
        $projects[] = $line;
      }
      $i++;
    }
    fclose($file);

    // Set up columns for the table.
    $columns = [
      'name'    => 'Name',
      'result'  => 'Result',
      'error'   => 'Error',
      'type'    => 'Project type',
      'usage'   => 'Usage',
      'group'   => 'Group',
      'total'   => 'Total deprecations',
      'parsing' => 'Total parse errors',
      'now'     => 'Fix now',
      'nine'    => 'Once 8.7 unsupported',
      'na'      => 'Uncategorized',
      'plan'    => 'Drupal 9 plan',
    ];
    // Add sorting information to columns.
    $i = 0;
    foreach ($columns as $name => &$column) {
      $column = ['data' => $column, 'field' => $i];
      $i++;
    }
    // Sort by usage descending initially.
    $columns['usage']['sort'] = 'desc';

    // Sort the data array based on query tablesort information.
    $order = tablesort_get_order($columns);
    $this->column = $order['sql'];
    $this->sort = tablesort_get_sort($columns);
    uasort($projects, [$this, 'tableSort']);

    $common_script = <<<SCRIPTEND
    responsive: false,
    legend: {
      display: true,
      position: 'right'
    },
    tooltips: {
      callbacks: {
        label: function(tooltipItem, data) {
          var dataset = data.datasets[tooltipItem.datasetIndex];
          var meta = dataset._meta[Object.keys(dataset._meta)[0]];
          var total = meta.total;
          var currentValue = dataset.data[tooltipItem.index];
          var percentage = parseFloat((currentValue/total*100).toFixed(1));
          return currentValue + ' (' + percentage + '%)';
        },
        title: function(tooltipItem, data) {
          return data.labels[tooltipItem[0].index];
        }
      }
    }
SCRIPTEND;
    $form['chart'] = [
      '#children' => <<<SCRIPTEND
<script src="https://cdn.jsdelivr.net/npm/chart.js@2.8.0"></script>
<div id="charts">
 <canvas id="totalProjects"></canvas>
 <canvas id="totalDeprecations"></canvas>
</div>
<script type="text/javascript">
var ctx = document.getElementById('totalProjects').getContext('2d');
var chart = new Chart(ctx, {
  type: 'doughnut',
  data: {
    labels: ['No problems found', 'All problems fixable now', 'Some problems fixable now', 'Needs manual review', 'Only problems fixable later', 'No results'],
    datasets: [{
      data: [@totalProjectsData],
      backgroundColor: ["#6ebe9f", "#2586a4", "#f3a935", "#55596a", "#c73558"]
    }]
  },
  options: {
    title: {
      display: true,
      text: '@projectcount projects by results',
      position: 'bottom'
    },
    $common_script
  }
});
var ctx = document.getElementById('totalDeprecations').getContext('2d');
var chart = new Chart(ctx, {
  type: 'doughnut',
  data: {
    labels: ['Fix now', 'Once 8.7 unsupported', 'Uncategorized'],
    datasets: [{
      data: [@totalDeprecationsData],
      backgroundColor: ["#f3a935", "#c73558"]
    }]
  },
  options: {
    title: {
      display: true,
      text: '@deprecatedcount deprecated API use',
      position: 'bottom'
    },
    $common_script
  }
});
</script>
SCRIPTEND
    ];

    // Display the resulting data.
    $form['table'] = [
      '#type' => 'table',
      '#header' => $columns,
      '#empty' => t('No matching projects found.'),
    ];

    // Filter projects down based on form input. Add data summaries of overall
    // results and the filtered results.
    $project_count = count($projects);
    $summary = $this->dataSummary($projects);
    $projects = $this->tableFilter($projects);
    if (count($projects) != $project_count) {
      $summary = $this->dataSummary($projects);
    }
    if (count($projects) == 0) {
      unset($form['chart']);
    }
    else {
      $project_data = join(',', array_values($summary['results']));
      $deprecation_data = join(',', array_values($summary['errors']));
      $deprecation_count = $summary['total'];
      $form['chart']['#children'] = str_replace(
        ['@totalProjectsData', '@totalDeprecationsData', '@projectcount', '@deprecatedcount'],
        [$project_data, $deprecation_data, count($projects), $deprecation_count],
        $form['chart']['#children']
      );
    }

    // Page the results based on query information.
    $page = pager_find_page();
    $num_per_page = 30;
    pager_default_initialize(count($projects), $num_per_page);
    $offset = $num_per_page * $page;
    $project_page = array_slice($projects, $offset, $num_per_page);

    // Get dataset metadata for outside links to error logs.
    list ($data_id, $data_date) = DataSource::getMetaData();

    // Key results based on header key names.
    $keys = array_keys($columns);
    foreach ($project_page as $project) {
      $info = [];
      if ($project[1] == 'No results') {
        $log_uri =
          'https://dispatcher.drupalci.org/job/phpstan/' . $data_id .
          '/artifact/phpstan-results/' . str_replace(' ', '.', $project[0]) . '.phpstan_stderr/*view*/';
        $info['name'] = [
          '#type' => 'markup',
          '#markup' => $project[0],
        ];
        $info['result'] = [
          'data' => [
            '#type' => 'link',
            '#title' => 'No results',
            '#url' => Url::fromUri($log_uri),
            '#attributes' => ['class' => ['category']],
          ],
        ];
        $class = 'unknown';
      }
      else {
        $info['name'] = [
          'data' => [
            '#type' => 'link',
            '#title' => $project[0],
            '#url' => Url::fromRoute('deprecation_status.errors_form', [], ['query' => ['project' => $project[0]]])
          ]
        ];
        $info['result'] = [
          '#type' => 'markup',
          '#markup' => '<span class="category">' . $project[1] . '</span>',
        ];
        $class = 'passed';
        if (in_array($project[1], ['All problems fixable now', 'Some problems fixable now', 'Needs manual review', 'Only problems fixable later'])) {
          $class = 'warning';
        }
      }
      $info['#attributes']['class'] = ['item-' . $class];

      foreach($project as $index => $item) {
        if (!isset($info[$keys[$index]])) {
          $info[$keys[$index]] = [
            '#type' => 'markup',
            '#markup' => $item,
          ];
        }
      }
      // Format drupal.org issue links in plans and fix relative links.
      $plan = preg_replace('!(\[#([0-9]+)\])!', '<a target="_blank" href="https://drupal.org/node/\2">\1</a>', $project[11]);
      $plan = preg_replace('!href="/!', 'href="https://drupal.org/', $plan);
      $info['plan'] = [
        '#type' => 'markup',
        '#markup' => $plan,
      ];
      $info['now']['#markup'] = '<strong>' . $info['now']['#markup'] . '</strong>';

      $form['table'][] = $info;
    }

    $form['pager'] = array(
      '#type' => 'pager'
    );

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {}

  /**
   * Custom table sorting based on stored sorting criteria.
   */
  protected function tableSort($a, $b) {
    if ($a[$this->column] == $b[$this->column]) {
      return 0;
    }
    $cmp = ($a[$this->column] < $b[$this->column]) ? -1 : 1;
    return ($this->sort == 'desc') ? $cmp * -1 : $cmp;
  }

  /**
   * Custom table filter based on form values.
   */
  protected function tableFilter($projects) {
    $keys = ['name', 'result', 'error', 'type', 'group', 'plan'];
    foreach ($keys as $key) {
      $$key = @$_GET[$key];
    }

    $filtered = [];
    foreach ($projects as $project) {
      // Assume this row is a match.
      $match = TRUE;

      // Go through each filter and if the row is still a match, see if this
      // filter will invalidate it.
      if (!empty($name) && $match) {
        $match = (strpos($project[0], $name) !== FALSE);
      }
      if (!empty($result) && $match) {
        $match = ($project[1] == $result);
      }
      if (!empty($error) && $match) {
        $match = ($project[2] == $error);
      }
      if (!empty($type) && $match) {
        $match = ($project[3] == $type);
      }
      if (!empty($group) && $match) {
        $match = ($project[5] <= $group);
      }
      if (!empty($plan) && $match) {
        $match = ($plan == 'none' ? empty($project[11]) : !empty($project[11]));
      }

      // If the row was still a match, continue.
      if ($match) {
        $filtered[] = $project;
      }
    }

    return $filtered;
  }

  protected function dataSummary($projects) {
    $summary = [
      'projects' => count($projects),
      'total' => 0,
      'errors' => [
        'now' => 0,
        'nine' => 0,
        'na' => 0,
      ],
      'results' => [
        'No problems found' => 0,
        'All problems fixable now' => 0,
        'Some problems fixable now' => 0,
        'Needs manual review' => 0,
        'Only problems fixable later' => 0,
        'No results' => 0,
      ],
    ];
    foreach($projects as $project) {
      $summary['results'][$project[1]]++;
      $summary['total'] += $project[6];
      $summary['errors']['now'] += $project[8];
      $summary['errors']['nine'] += $project[9];
      $summary['errors']['na'] += $project[10];
    }
    return $summary;
  }

}
